#!/usr/bin/ruby
#
#   check-query-json
#
# DESCRIPTION:
#   Based on the query that you are trying to execute this verify
#   if the query satisfy the requirements on the thresholds
#   You can specify a column and warning and error expressions
#
# OUTPUT:
#   plain text
#
# PLATFORMS:
#   Linux
#
# DEPENDENCIES:
#   gem: sensu-plugin
#   gem: json
#
# USAGE:
#   #YELLOW
#
# NOTES:
#   Based on Check HTTP.
#
# LICENSE:
#   Copyright 2017 Ve Interactive <velabs@veinteractive.com>
#   Released under the Apache 2.0 license; see LICENSE
#   for details.
#

require 'sensu-plugin/check/cli'
require 'json'
require 'net/http'
require 'net/https'
require 'open-uri'

#
# Check query
#
class QueryCheck < Sensu::Plugin::Check::CLI
  option :apiUrl,
         short: '-a URL',
         long: '--apiURL URL',
         default: 'https://api.applicationinsights.io/beta'
  option :apiVersion,
         short: '-V beta',
         long: '--apiVersion beta',
         default: 'beta'
  option :appId,
         short: '-i GUID',
         long: '--appId GUID',
         required: true
  option :appKey,
         short: '-p VALUE',
         long: '--appKey VALUE',
         required: true
  option :query,
         short: '-q /file/with/query or https://fileUrl',
         long: '--query /file/with/query or https://fileUrl',
         description: 'Query can be expressed as a path to a file a url that will return the raw query or directly inline.',
         required: true
  option :resultColumn,
         short: '-r COLUMN_NAME',
         long: '--resultColumn COLUMN_NAME'
  option :warningExpression,
         short: '-w WARNING_EXPRESSION value > 20',
         long: '--warning WARNING_EXPRESSION value > 20'
  option :criticalExpression,
         short: '-c CRITICAL_EXPRESSION value > 20',
         long: '--critical CRITICAL_EXPRESSION value > 20'
  option :message,
         short: '-m MESSAGE',
         long: '--message MESSAGE'
  option :timeout,
         short: '-t SECS',
         long: '--timeout SECS',
         proc: proc(&:to_i),
         default: 60

  def run
    uri = URI.parse(config[:apiUrl])
    config[:host] = uri.host
    config[:path] = uri.path
    config[:port] = uri.port
    config[:ssl] = uri.scheme == 'https'
    config[:port] ||= config[:ssl] ? 443 : 80
    begin
      Timeout.timeout(config[:timeout]) do
        validate_query
      end
    rescue Timeout::Error
      critical 'Query timed out'
    rescue => e
      critical "Error:executing #{e.message}"
    end
  end

  def json_valid?(str)
    JSON.parse(str)
    return true
  rescue JSON::ParserError
    return false
  end

  def validate_query
    http = Net::HTTP.new(config[:host], config[:port])

    if config[:ssl]
      http.use_ssl = true
      http.verify_mode = OpenSSL::SSL::VERIFY_NONE
    end
    query = get_query(config[:query])
    encoded = URI.encode(query)
    query_url = [[config[:path], 'apps', config[:appId], 'query?query='].compact.join('/'), encoded].compact.join
    req = Net::HTTP::Get.new(query_url)
    req['x-api-key'] = config[:appKey]
    res = http.request(req)
    unless res.code =~ /^2/
      critical "http code: #{res.code}: body: #{res.body}"
    end
    critical 'invalid JSON from request' unless json_valid?(res.body)
    begin
      json = JSON.parse(res.body)
    rescue JSON::ParserError
      critical 'Not able to parse response from api'
    end
    validate_result(json, query)
  end

  def get_query(query)
    post_body = ''
    if query =~ /\A#{URI.regexp(%w(https http))}\z/
      uri = URI.parse(query)
      http_query = Net::HTTP.new(uri.host, uri.port)
      if uri.scheme == 'https'
        http_query.use_ssl = true
        http_query.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end
      req_query = Net::HTTP::Get.new(uri.path)
      output "request query starting #{req_query}, #{query}"
      res_query = http_query.get(req_query)
      output 'request query'
      unless res_query.code =~ /^2/
        critical "http code: #{res_query.code}: body: #{res_query.body}"
      end
      post_body = res_query.body
    elsif File.exist?(query)
      post_body = IO.readlines(query).join
    else
      post_body = query
    end
    post_body
  end

  def validate_result(json, query)
    columns = json['Tables'][0]['Columns']
    rows = json['Tables'][0]['Rows']
    column_index = get_column_index(columns)
    critical_messages = []
    warning_message = []
    rows.each_with_index do |row, _index|
      values = extract_value(row, columns)
      value = row[column_index]
      if values.nil? || value.nil?
        critical 'unable to extract values'
      end

      if config[:warningExpression]
        if instance_eval config[:warningExpression]
          if config[:message]
            warning_message.push instance_eval config[:message]
          else
            warning_message.push "#{row[column_index]} it is #{config[:warningExpression]}"
          end
        end
      end
      if config[:criticalExpression]
        if instance_eval config[:criticalExpression]
          if config[:message]
            critical_messages.push instance_eval config[:message]
          else
            critical_messages.push "#{row[column_index]} it is #{config[:criticalExpression]}"
          end
        end
      end
    end
    message "\nQuery:#{query} \n Criticals: #{critical_messages} \n Warnings: #{warning_message}"
    critical unless critical_messages.empty?
    warning unless warning_message.empty?
    ok
  end

  def get_column_index(columns)
    column_index = -1
    columns.each_with_index do |value, index|
      if value['ColumnName'] == config[:resultColumn]
        column_index = index
      elsif column_index == -1 && %w(real long).include?(value['ColumnType'])
        column_index = index
      end
    end
    if column_index == -1
      column_index = 0
    end
    column_index
  end

  def extract_value(row, columns)
    result = {}
    columns.each_with_index do |value, index|
      name = value['ColumnName']
      result[name] = row[index]
    end
    result
  end
end
